# Copyright (c) AIRBUS and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import time
from abc import abstractmethod
from datetime import timedelta
from typing import Any, Iterable, Optional

from skdecide.discrete_optimization.generic_tools.cp_tools import CPSolver, ParametersCP
from skdecide.discrete_optimization.generic_tools.do_problem import (
    ModeOptim,
    ParamsObjectiveFunction,
    Problem,
    Solution,
    build_aggreg_function_and_params_objective,
    build_evaluate_function_aggregated,
)
from skdecide.discrete_optimization.generic_tools.do_solver import SolverDO
from skdecide.discrete_optimization.generic_tools.lns_mip import (
    InitialSolution,
    PostProcessSolution,
    TrivialPostProcessSolution,
)
from skdecide.discrete_optimization.generic_tools.result_storage.result_storage import (
    ResultStorage,
)


class ConstraintHandler:
    @abstractmethod
    def adding_constraint_from_results_store(
        self, cp_solver: CPSolver, child_instance, result_storage: ResultStorage
    ) -> Iterable[Any]:
        ...

    @abstractmethod
    def remove_constraints_from_previous_iteration(
        self, cp_solver: CPSolver, child_instance, previous_constraints: Iterable[Any]
    ):
        ...


class LNS_CP(SolverDO):
    def __init__(
        self,
        problem: Problem,
        cp_solver: CPSolver,
        initial_solution_provider: InitialSolution,
        constraint_handler: ConstraintHandler,
        post_process_solution: PostProcessSolution = None,
        params_objective_function: ParamsObjectiveFunction = None,
    ):
        self.problem = problem
        self.cp_solver = cp_solver
        self.initial_solution_provider = initial_solution_provider
        self.constraint_handler = constraint_handler
        self.post_process_solution = post_process_solution
        if self.post_process_solution is None:
            self.post_process_solution = TrivialPostProcessSolution()
        self.params_objective_function = params_objective_function
        (
            self.aggreg_from_sol,
            self.aggreg_dict,
            self.params_objective_function,
        ) = build_aggreg_function_and_params_objective(
            problem=self.problem,
            params_objective_function=self.params_objective_function,
        )

    def solve_lns(
        self,
        parameters_cp: ParametersCP,
        nb_iteration_lns: int,
        nb_iteration_no_improvement: Optional[int] = None,
        max_time_seconds: Optional[int] = None,
        skip_first_iteration: bool = False,
        **args
    ) -> ResultStorage:
        sense = self.params_objective_function.sense_function
        if max_time_seconds is None:
            max_time_seconds = 3600 * 24  # One day
        if nb_iteration_no_improvement is None:
            nb_iteration_no_improvement = 2 * nb_iteration_lns
        current_nb_iteration_no_improvement = 0
        deb_time = time.time()
        if not skip_first_iteration:
            store_lns = self.initial_solution_provider.get_starting_solution()
            init_solution, objective = store_lns.get_best_solution_fit()
            best_solution = init_solution.copy()

            satisfy = self.problem.satisfy(init_solution)
            best_objective = objective
        else:
            best_objective = (
                float("inf") if sense == ModeOptim.MINIMIZATION else -float("inf")
            )
            best_solution = None
            constraint_iterable = {"empty": []}
            store_lns = None
        for iteration in range(nb_iteration_lns):
            with self.cp_solver.instance.branch() as child:
                if iteration == 0 and not skip_first_iteration or iteration >= 1:
                    constraint_iterable = (
                        self.constraint_handler.adding_constraint_from_results_store(
                            cp_solver=self.cp_solver,
                            child_instance=child,
                            result_storage=store_lns,
                        )
                    )
                try:
                    if iteration == 0:
                        result = child.solve(
                            timeout=timedelta(seconds=parameters_cp.TimeLimit_iter0),
                            intermediate_solutions=parameters_cp.intermediate_solution,
                        )
                    else:
                        result = child.solve(
                            timeout=timedelta(seconds=parameters_cp.TimeLimit),
                            intermediate_solutions=parameters_cp.intermediate_solution,
                        )
                    result_store = self.cp_solver.retrieve_solutions(
                        result, parameters_cp=parameters_cp
                    )
                    if len(result_store.list_solution_fits) > 0:
                        bsol, fit = result_store.get_best_solution_fit()
                        result_store = self.post_process_solution.build_other_solution(
                            result_store
                        )
                        bsol, fit = result_store.get_best_solution_fit()
                        if sense == ModeOptim.MAXIMIZATION and fit >= best_objective:
                            if fit > best_objective:
                                current_nb_iteration_no_improvement = 0
                            else:
                                current_nb_iteration_no_improvement += 1
                            best_solution = bsol
                            best_objective = fit
                        elif sense == ModeOptim.MAXIMIZATION:
                            current_nb_iteration_no_improvement += 1
                        elif sense == ModeOptim.MINIMIZATION and fit <= best_objective:
                            if fit < best_objective:
                                current_nb_iteration_no_improvement = 0
                            else:
                                current_nb_iteration_no_improvement += 1
                            best_solution = bsol
                            best_objective = fit
                        elif sense == ModeOptim.MINIMIZATION:
                            current_nb_iteration_no_improvement += 1
                        if skip_first_iteration and iteration == 0:
                            store_lns = result_store
                        for s, f in result_store.list_solution_fits:
                            store_lns.add_solution(solution=s, fitness=f)
                    else:
                        current_nb_iteration_no_improvement += 1
                except Exception as e:
                    current_nb_iteration_no_improvement += 1
                if time.time() - deb_time > max_time_seconds:
                    break
                if current_nb_iteration_no_improvement > nb_iteration_no_improvement:
                    break

        return store_lns
